/*
	bootparam_default_route.c
	jelinekr@ms.com		01/18/05
*/

	
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <asm/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <netdb.h>
#include <syslog.h>
#include <arpa/inet.h>
#include <unistd.h>

#define MAXNREQS	5
#define BUFSIZE		8192
#define PATHSIZE	64
#define PROCIPFWDPATH	"/proc/sys/net/ipv4/conf"

extern	int		debug;
extern	int		dolog;

u_long                  get_router(struct in_addr src_addr);
u_long			get_default_route(struct in_addr);
void			msgout(char *, ...);

/*
 * Make the following attempts to find a router for the client
 * 1. find a router by sending ICMP_ECHO to ALL_ROUTERS
 * 2. return our own interface address if we're a router
 * 3  see if our default route can be used by the client
 */
u_long
get_default_route(cl_addr)
	struct in_addr cl_addr;			/* client address */
{
	int			n;
	int			sd;
	char			proc_ipfwpath[PATHSIZE];
	struct in_addr		if_addr;	/* interface address */
	struct in_addr		if_mask;	/* interface mask */
	struct in_addr		rt_addr;	/* router address */

	if_addr.s_addr = 0L;
	if_mask.s_addr = 0L;
	rt_addr.s_addr = 0L;


	if (debug)
		msgout("cl_addr = %s", inet_ntoa(cl_addr));


	/*
	 * Get list of interfaces
	 */

	{
		struct ifconf		ifc;
		struct ifreq		ifr;
		struct ifreq		*ifrp;
		int			numreqs = MAXNREQS;
		
	
	
		if ((sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0) {
			perror("socket");
			return ((u_long)0);
		}

		ifc.ifc_buf = NULL;

		for (;;) {

			ifc.ifc_len = sizeof(struct ifreq) * numreqs;

			if ((ifc.ifc_buf = 
				realloc(ifc.ifc_buf, ifc.ifc_len)) == NULL) {
				perror("realloc");
				return ((u_long)0);
			}

			if (ioctl(sd, SIOCGIFCONF, &ifc) < 0) {
				perror("SIOCGIFCONF");
				free(ifc.ifc_buf);
				close(sd);
				return ((u_long)0);
			}

			if (ifc.ifc_len == (int)(sizeof(struct ifreq) * numreqs)) {
				if (debug)
					msgout("Need to realloc() (%d)", numreqs);
				numreqs += 5;
				continue;
			}

			break;
		}

		ifrp = ifc.ifc_req;

		/*
		 *  Find out which interface is on client's subnet
		 */

		for (n = 0; n < ifc.ifc_len; n += sizeof(struct ifreq)) {

			struct sockaddr_in addr;
			struct sockaddr_in mask;
			struct sockaddr sa;

			memset(&ifr, 0, sizeof(struct ifreq));

			strncpy(ifr.ifr_name, ifrp->ifr_name, IFNAMSIZ);

			if (ioctl(sd, SIOCGIFADDR, &ifr) != 0) {
				continue;
			} else {
				sa = ifr.ifr_addr;
				memcpy(&addr, (struct sockaddr_in *)&sa, 
					sizeof(struct sockaddr_in));
			}


			if (ioctl(sd, SIOCGIFNETMASK, &ifr) < 0) {
				continue;
			} else {
				sa = ifr.ifr_addr; /* ifr_netmask; */
				memcpy(&mask, (struct sockaddr_in *)&sa, 
					sizeof(struct sockaddr_in));
			}

			if (debug)
				msgout("if_name = %s, if_addr = %x , if_mask = %x",
					ifr.ifr_name, 
					ntohl(addr.sin_addr.s_addr),
					ntohl(mask.sin_addr.s_addr));

			/* assuming that netmask is correctly set */
			if ((addr.sin_addr.s_addr & mask.sin_addr.s_addr) == 
				(cl_addr.s_addr & mask.sin_addr.s_addr)) {
				sprintf(proc_ipfwpath, "%s/%s/forwarding",
					PROCIPFWDPATH, ifr.ifr_name);
				memcpy(&if_addr, &addr.sin_addr, 
					sizeof(struct in_addr));
				memcpy(&if_mask, &mask.sin_addr, 
					sizeof(struct in_addr));

			}
			ifrp++;
		}
		close(sd);
		free(ifc.ifc_buf);

		if (if_addr.s_addr == (u_long)0) {
			if (debug)
				msgout("no interface on client's subnet");
			return ((u_long)0);
		}
	}


	/*
	 * Try to find a router by sending ICMP_ECHO to ALL_ROUTERS multicast group
	 */

	if ((rt_addr.s_addr = (u_long)get_router(if_addr)) != 0) {
		if (debug)
			msgout("rt_addr = %s", inet_ntoa(rt_addr));
		return(rt_addr.s_addr);
	}


	/*
	 * If ip forwarding is enabled, return ip address of the interface that's on 
	 * client's subnet
	 */

	{
		FILE		*fd;
		char		buf[2];
		int		ip_fw = 0;

		if ((fd = fopen(proc_ipfwpath, "r")) != NULL) {
			fgets(buf, sizeof(buf), fd);
			ip_fw = atoi(buf);

			(void) fclose(fd);
		}
	
		if (ip_fw == 1) {
			if (debug)
				msgout("rt_addr = %s, ip_fw = %d",
					inet_ntoa(if_addr), ip_fw);
			return(if_addr.s_addr);
		}
	}


	/*
	 * See if our default route could be used by the client
	 */

	{
		struct nlmsghdr		*nlMsg;
		struct rtmsg		*rtMsg;
		char			buf[BUFSIZE];
		int			len = 0;
		int			sd;

		if ((sd = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE)) < 0) {
			perror("socket");
			return ((u_long)0);
		}

		memset(buf, 0, BUFSIZE);

		nlMsg = (struct nlmsghdr *)buf;
		rtMsg = (struct rtmsg *)NLMSG_DATA(nlMsg);

        	nlMsg->nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
		nlMsg->nlmsg_type = RTM_GETROUTE;
		nlMsg->nlmsg_flags = NLM_F_DUMP | NLM_F_REQUEST;
		nlMsg->nlmsg_seq = 0;             
		nlMsg->nlmsg_pid = getpid();

		if (send(sd, nlMsg, nlMsg->nlmsg_len, 0) < 0){
			perror("send");
			return ((u_long)0);
		}

		{
			struct nlmsghdr		*nlHdr;
			int			nbytes = 0;
			char			*bufp = buf;

			do {
	
				if ((nbytes = recv(sd, bufp, BUFSIZE - len, 0)) < 0) {
					perror("recv");
					return((u_long)0);
				}

				nlHdr = (struct nlmsghdr *)bufp;

				if ((NLMSG_OK(nlHdr, (u_int)nbytes) == (u_int)0) || 
					(nlHdr->nlmsg_type == NLMSG_ERROR)) {
					perror("recv");
					return((u_long)0);
				}

				if (nlHdr->nlmsg_type == NLMSG_DONE)
					break;
				else {
					bufp += nbytes;
					len += nbytes;
				}

				if ((nlHdr->nlmsg_flags & NLM_F_MULTI) == 0)
					break;

			} while (nlHdr->nlmsg_seq != 0 ||
					nlHdr->nlmsg_pid != (u_int)getpid());

		}

		if (len == 0)
			return((u_long)0);

		if (debug)
			msgout("Read total %d bytes", len);


		for (;NLMSG_OK(nlMsg,(u_int)len);nlMsg = NLMSG_NEXT(nlMsg,len)) {

			struct rtattr		*rtAttr;
			struct rtmsg		*rtMsg;
			int			rtLen;
			struct route_info	{
				u_int	dst;
				u_int	src;
				u_int	gw;
				char	ifName[IF_NAMESIZE];
			} rt;
			
			memset(&rt, 0, sizeof(struct route_info));

			rtMsg = (struct rtmsg *)NLMSG_DATA(nlMsg);

			if ((rtMsg->rtm_family != AF_INET) ||
				(rtMsg->rtm_table != RT_TABLE_MAIN))
				continue;

			rtAttr = (struct rtattr *)RTM_RTA(rtMsg);
			rtLen = RTM_PAYLOAD(nlMsg);

			for (;RTA_OK(rtAttr,rtLen);rtAttr = RTA_NEXT(rtAttr,rtLen)) {

				if (rtAttr->rta_type == RTA_OIF) {
					if_indextoname(*(int *)RTA_DATA(rtAttr), rt.ifName);
				} else if (rtAttr->rta_type == RTA_GATEWAY) {
					rt.gw = *(u_int *)RTA_DATA(rtAttr);
				} else if (rtAttr->rta_type == RTA_PREFSRC) {
					rt.src = *(u_int *)RTA_DATA(rtAttr);
				} else if (rtAttr->rta_type == RTA_DST) {
					rt.dst = *(u_int *)RTA_DATA(rtAttr);
				}
			}

			if (debug)
				msgout("if = %s, dest = %x, gw = %x",
					rt.ifName, ntohl(rt.dst), ntohl(rt.gw));

			/*
			 * We only want a default route that's on client's subnet 
			 */
			if ((rt.gw & if_mask.s_addr) !=
                                    (if_addr.s_addr & if_mask.s_addr))
                                        continue;

			if ((rt.dst == (u_int)0) ||
                                    (rt_addr.s_addr == (u_long)0)) {
                                        rt_addr.s_addr = (u_long)rt.gw;

					if (debug)
						msgout("rt_addr = %s, gw = %x",
							inet_ntoa(rt_addr), ntohl(rt.gw));
			}
                }

		if (rt_addr.s_addr == (u_long)0)
			if (debug)
				msgout("no default route found");

		return (rt_addr.s_addr);
	}
}


#ifdef  __STDC__
#include <stdarg.h>

void
msgout(char *fmt, ...)
{
	va_list	args;
	va_start(args, fmt);

#else
#include <varargs.h>

void msgout(fmt, va_alist)
	char		*fmt;
	va_dcl
{
	va_list	args;
	va_start(args);

#endif /* __STDC__ */

	if (dolog)
		syslog(LOG_ERR, fmt, args);
	else {
		(void) fflush(stderr);
		(void) vfprintf(stderr, fmt, args);
		(void) fprintf(stderr, "\n");	
	}
	va_end(args);
}
